using System;
using System.Collections;
using System.Diagnostics;
using Goods;


[Metaobject(typeof(OptimisticMetaobject))]
public class Record:Persistent
{
    Record(Metaobject mop) : base(mop) {}

    public Record(long intKey, string strKey) 
    { 
        this.intKey = intKey;
        this.strKey = strKey;
    }

    public string strKey;
    public long   intKey;
}


[Metaobject(typeof(OptimisticMetaobject))]
public class Root:Persistent
{
    Root(Metaobject mop) : base(mop) {}

    public Root(SortedCollection intIndex, SortedCollection strIndex) 
    { 
        this.intIndex = intIndex;
        this.strIndex = strIndex;
    }

    public void initialize(long key, int nRecords) 
    { 
        for (int i = 0; i < nRecords; i++)
        {
            key = (3141592621L * key + 2718281829L) % 1000000007L;
            Record rec = new Record(key, Convert.ToString(key));
            intIndex.Add(rec);
            strIndex.Add(rec);
        }
    }

    public void remove(long key, int nRecords) 
    {
        for (int i = 0; i < nRecords; i++)
        {
            key = (3141592621L * key + 2718281829L) % 1000000007L;
            Record rec = (Record) intIndex[key];
            intIndex.Remove(rec);
            strIndex.Remove(rec);
            //rec.Deallocate();
        }
    }

    [Constant]
    public void check(long key, int nRecords) 
    {
        key = 1999;
        for (int i = 0; i < nRecords; i++)
        {
            key = (3141592621L * key + 2718281829L) % 1000000007L;
            Record rec1 = (Record) intIndex[key];
            Record rec2 = (Record) strIndex[Convert.ToString(key)];
            Debug.Assert(rec1 != null && rec1 == rec2);
        }
    }

    [Constant]
    public void iterate(int nRecords) 
    {
        long key = Int64.MinValue;
        int i = 0;
        foreach (Record rec in intIndex) 
        {
            Debug.Assert(rec.intKey >= key);
            key = rec.intKey;
            i += 1;
        }
        Debug.Assert(i == nRecords);
        i = 0;
        String strKey = "";
        foreach (Record rec in strIndex) 
        {
            Debug.Assert(rec.strKey.CompareTo(strKey) >= 0);
            strKey = rec.strKey;
            i += 1;
        }
        Debug.Assert(i == nRecords);
    }

    SortedCollection strIndex;
    SortedCollection intIndex;
}

[Metaobject(typeof(OptimisticMetaobject))]
public class RecordIntKeyComparator:PersistentComparator
{
    RecordIntKeyComparator(Metaobject mop) : base(mop) {}

    public RecordIntKeyComparator() {}

    [Constant]
    public override int CompareMembers(Persistent m1, Persistent m2) 
    {
        long diff = ((Record)m1).intKey - ((Record)m2).intKey;
        return diff < 0 ? -1 : diff == 0 ? 0 : 1;
    }

    [Constant]
    public override int CompareMemberWithKey(Persistent mbr, object key) 
    {
        long diff = ((Record)mbr).intKey - (long)key;
        return diff < 0 ? -1 : diff == 0 ? 0 : 1;
    }
}

[Metaobject(typeof(OptimisticMetaobject))]
public class RecordStrKeyComparator:PersistentComparator
{
    RecordStrKeyComparator(Metaobject mop) : base(mop) {}

    public RecordStrKeyComparator() {}

    [Constant]
    public override int CompareMembers(Persistent m1, Persistent m2) 
    {
        return ((Record)m1).strKey.CompareTo(((Record)m2).strKey);
    }

    [Constant]
    public override int CompareMemberWithKey(Persistent mbr, object key) 
    {
        return ((Record)mbr).strKey.CompareTo(key);
    }
}



public class TestIndex
{
    internal const int nRecords = 10000;
	
    static public void Main()
    {
        Database db = new Database();
        if (!db.Open("testidx.cfg")) 
        {
            Console.WriteLine("Failed to connect to server");
            return;
        }
        Root root = (Root) db.Root;
        if (root == null)
        {
            root = new Root(new SortedCollection(new RecordIntKeyComparator(), true),
                            new SortedCollection(new RecordStrKeyComparator(), true));
            db.Root = root;
        }
        DateTime start = DateTime.Now;
        long key = 1999;
        root.initialize(key, nRecords);
        System.Console.WriteLine("Elapsed time for inserting " + nRecords + " records: " + (DateTime.Now - start));
		
        start = System.DateTime.Now;
        root.check(key, nRecords);
        System.Console.WriteLine("Elapsed time for performing " + nRecords * 2 + " index searches: " + (DateTime.Now - start));

        start = System.DateTime.Now;
        root.iterate(nRecords);
        System.Console.WriteLine("Elapsed time for iteration through " + (nRecords * 2) + " records: " + (DateTime.Now - start));

        start = System.DateTime.Now;
        root.remove(key, nRecords);
        System.Console.WriteLine("Elapsed time for deleting " + nRecords + " records: " + (DateTime.Now - start));

        db.Close();
    }
}