//-< FieldDescriptor.java >------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 13-Jan-99    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Class reponsible for packing/unpacking fields of persistent object
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    class FieldDescriptor
    {
        internal int appType;
        internal int dbsType;
        internal int size;
        internal String name;
        internal System.Reflection.FieldInfo field;
        internal Type cls;
        internal bool isVarying;
		
        internal FieldDescriptor next;
        internal FieldDescriptor nextRef;
        internal FieldDescriptor nextStr;
        internal FieldDescriptor nextBin;
		
        internal const int t_sbyte = 0;
        internal const int t_short = 1;
        internal const int t_integer = 2;
        internal const int t_long = 3;
        internal const int t_ubyte = 4;
        internal const int t_ushort = 5;
        internal const int t_uinteger = 6;
        internal const int t_ulong = 7;
        internal const int t_float = 8;
        internal const int t_double = 9;
        internal const int t_boolean = 10;
        internal const int t_char = 11;
        internal const int t_reference = 12;
        internal const int t_string = 13;
        internal const int t_date = 14;
        internal const int t_raw_binary = 15;
		
        internal const int nullString = 0xFFFF;
		
        internal FieldDescriptor(System.Reflection.FieldInfo f, Type cls)
        {
            Type fieldClass = f.FieldType;
            this.name = f.Name;
            this.cls = cls;
            this.field = f;
            if (typeof(Persistent).IsAssignableFrom(fieldClass) || fieldClass.IsInterface)
            {
                appType = t_reference;
                dbsType = Protocol.fld_reference;
                size = Protocol.ref_size;
            }
            else
            {
                if (fieldClass == typeof(byte) || fieldClass == typeof(byte[])) 
                { 
                    dbsType = (byte)Protocol.fld_unsigned_integer;
                    appType = t_ubyte;
                    size = 1;
                } 
                else if (fieldClass == typeof(sbyte) || fieldClass == typeof(sbyte[])) 
                { 
                    dbsType = (byte)Protocol.fld_signed_integer;
                    appType = t_sbyte;
                    size = 1;
                } 
                else if (fieldClass == typeof(short) || fieldClass == typeof(short[]))
                { 
                    dbsType = (byte)Protocol.fld_signed_integer;
                    appType = t_short;
                    size = 2;
                } 
                else if (fieldClass == typeof(ushort) || fieldClass == typeof(ushort[]))
                { 
                    dbsType = (byte)Protocol.fld_unsigned_integer;
                    appType = t_ushort;
                    size = 2;
                } 
                else if (fieldClass == typeof(int) || fieldClass == typeof(int[])) 
                { 
                    dbsType = (byte)Protocol.fld_signed_integer;
                    appType = t_integer;
                    size = 4;
                } 
                else if (fieldClass == typeof(uint) || fieldClass == typeof(uint[])) 
                { 
                    dbsType = (byte)Protocol.fld_unsigned_integer;
                    appType = t_uinteger;
                    size = 4;
                } 
                else if (fieldClass == typeof(long) || fieldClass == typeof(long[])) 
                { 
                    dbsType = (byte)Protocol.fld_signed_integer;
                    appType = t_long;
                    size = 8;
                } 
                else if (fieldClass == typeof(ulong) || fieldClass == typeof(ulong[])) 
                { 
                    dbsType = (byte)Protocol.fld_unsigned_integer;
                    appType = t_ulong;
                    size = 8;
                } 
                else if (fieldClass == typeof(float) || fieldClass == typeof(float[])) 
                { 
                    dbsType = (byte)Protocol.fld_real;
                    appType = t_float;
                    size = 4;
                } 
                else if (fieldClass == typeof(double) || fieldClass == typeof(double[])) 
                { 
                    dbsType = (byte)Protocol.fld_real;
                    appType = t_double;
                    size = 8;
                } 
                else if (fieldClass == typeof(bool) || fieldClass == typeof(bool[])) 
                { 
                    dbsType = (byte)Protocol.fld_unsigned_integer;
                    appType = t_boolean;
                    size = 1;
                } 
                else if (fieldClass == typeof(char) || fieldClass == typeof(char[])) 
                { 
                    dbsType = (byte)Protocol.fld_unsigned_integer;
                    appType = t_char;
                    size = 2;
                } 
                else if (fieldClass == typeof(string)) 
                { 
                    dbsType = (byte)Protocol.fld_string;
                    appType = t_string;
                    size = 2;
                } 
                else if (fieldClass == typeof(DateTime)) 
                { 
                    dbsType = (byte)Protocol.fld_signed_integer;
                    appType = t_date;
                    size = 8;
                } 
                else
                {
                    Type componentType = fieldClass.GetElementType();
                    if (componentType != null && (typeof(Persistent).IsAssignableFrom(componentType) || componentType.IsInterface))
                    {
                        dbsType = Protocol.fld_reference;
                        appType = t_reference;
                        size = Protocol.ref_size;
                    }
                    else if (Database.rawBinaryFactory != null && Database.rawBinaryFactory.supports(fieldClass))
                    {
                        dbsType = Protocol.fld_raw_binary;
                        appType = t_raw_binary;
                        size = 4;
                    }
                    else
                    {
                        throw new ClassDefinitionError("Type of non-transient field '" + name + "'should be derived from Persistent class");
                    }
                }
                isVarying = appType != t_raw_binary && fieldClass.IsArray;
            }
        }
		
        internal FieldDescriptor(String name, Type cls, int dbsType, int size)
        {
            this.dbsType = dbsType;
            this.size = size;
            this.name = name;
            this.cls = cls;
            field = cls.GetField(name, System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.DeclaredOnly);
            if (field == null)
            {
                return;
            }
            Type fieldClass = field.FieldType;
            if (typeof(Persistent).IsAssignableFrom(fieldClass) || fieldClass.IsInterface)
            {
                if (dbsType != Protocol.fld_reference)
                {
                    throw new ClassDefinitionError("Field + '" + name + "' can not be converted to reference type");
                }
                appType = t_reference;
            }
            else
            {
                if (fieldClass == typeof(byte) || fieldClass == typeof(byte[])) 
                { 
                    appType = t_ubyte;
                } 
                else if (fieldClass == typeof(sbyte) || fieldClass == typeof(sbyte[])) 
                { 
                    appType = t_sbyte;
                } 
                else if (fieldClass == typeof(short) || fieldClass == typeof(short[]))
                { 
                    appType = t_short;
                } 
                else if (fieldClass == typeof(ushort) || fieldClass == typeof(ushort[]))
                { 
                    appType = t_ushort;
                } 
                else if (fieldClass == typeof(int) || fieldClass == typeof(int[])) 
                { 
                    appType = t_integer;
                } 
                else if (fieldClass == typeof(uint) || fieldClass == typeof(uint[])) 
                { 
                    appType = t_uinteger;
                } 
                else if (fieldClass == typeof(long) || fieldClass == typeof(long[])) 
                { 
                    appType = t_long;
                } 
                else if (fieldClass == typeof(ulong) || fieldClass == typeof(ulong[])) 
                { 
                    appType = t_ulong;
                } 
                else if (fieldClass == typeof(float) || fieldClass == typeof(float[])) 
                { 
                    appType = t_float;
                } 
                else if (fieldClass == typeof(double) || fieldClass == typeof(double[])) 
                { 
                    appType = t_double;
                } 
                else if (fieldClass == typeof(bool) || fieldClass == typeof(bool[])) 
                { 
                    appType = t_boolean;
                } 
                else if (fieldClass == typeof(char) || fieldClass == typeof(char[])) 
                { 
                    appType = t_char;
                } 
                else if (fieldClass == typeof(string)) 
                { 
                    appType = t_string;
                } 
                else if (fieldClass == typeof(DateTime)) 
                { 
                    appType = t_date;
                } 
                else
                {
                    Type componentType = fieldClass.GetElementType();
                    if (componentType != null && (typeof(Persistent).IsAssignableFrom(componentType) || componentType.IsInterface))
                    {
                        if (dbsType != Protocol.fld_reference)
                        {
                            throw new ClassDefinitionError("Field + '" + name + "' can not be converted to array of reference type");
                        }
                        appType = t_reference;
                    }
                    else if (Database.rawBinaryFactory != null && Database.rawBinaryFactory.supports(fieldClass))
                    {
                        if (dbsType != Protocol.fld_raw_binary)
                        {
                            throw new ClassDefinitionError("Field + '" + name + "' can not be converted to raw binary type");
                        }
                        appType = t_raw_binary;
                    }
                    else
                    {
                        throw new ClassDefinitionError("Type of non-transient field '" + name + "' should be derived from Persistent class");
                    }
                }
                isVarying = appType != t_raw_binary && fieldClass.IsArray;
            }
            if (appType != t_reference && dbsType == Protocol.fld_reference)
            {
                throw new ClassDefinitionError("Field + '" + name + "' can not be converted from reference type");
            }
            else if (appType != t_string && dbsType == Protocol.fld_string)
            {
                throw new ClassDefinitionError("Field + '" + name + "' can not be converted from string type");
            }
            else if (appType != t_raw_binary && dbsType == Protocol.fld_raw_binary)
            {
                throw new ClassDefinitionError("Field + '" + name + "' can not be converted from raw binary type");
            }
        }
		
        internal FieldDescriptor(FieldDescriptor copy)
        {
            // copy constructor
            appType = copy.appType;
            dbsType = copy.dbsType;
            size = copy.size;
            name = copy.name;
            field = copy.field;
            cls = copy.cls;
            isVarying = copy.isVarying;
        }
		
		
        internal int length(Persistent obj)
        {
            if (field != null)
            {
                Array arr = (Array)field.GetValue(obj);
                return arr != null ? arr.Length : 0;
            }
            return 0;
        }
		
        internal int pack(Persistent obj, byte[] data, int offs)
        {
            String s;
            switch (appType)
            {
                case t_sbyte: 
                case t_ubyte: 
                    data[offs] = (byte) field.GetValue(obj);
                    break;
				
                case t_short: 
                case t_ushort: 
                    Protocol.pack2(data, offs, (short) field.GetValue(obj));
                    break;
				
                case t_integer: 
                case t_uinteger: 
                    Protocol.pack4(data, offs, (int) field.GetValue(obj));
                    break;
				
                case t_long: 
                case t_ulong: 
                    Protocol.pack8(data, offs, (long) field.GetValue(obj));
                    break;
				
                case t_float: 
                    Protocol.pack4(data, offs,                         
                        BitConverter.ToInt32(BitConverter.GetBytes((float)field.GetValue(obj)), 0));
                    break;
				
                case t_double: 
                    Protocol.pack8(data, offs, 
                        BitConverter.DoubleToInt64Bits((double)field.GetValue(obj)));
                    break;

                case t_char: 
                    Protocol.pack2(data, offs, (char) field.GetValue(obj));
                    break;
				
                case t_boolean: 
                    data[offs] = (byte) ((bool) field.GetValue(obj)?1:0);
                    break;
				
                case t_string: 
                    s = (String) field.GetValue(obj);
                    if (s == null)
                    {
                        return Protocol.pack2(data, offs, nullString);
                    }
                    else
                    {
                        int len = s.Length;
                        offs = Protocol.pack2(data, offs, len);
                        for (int i = 0; i < len; i++)
                        {
                            offs = Protocol.pack2(data, offs, s[i]);
                        }
                        return offs;
                    }
				
                case t_raw_binary: 
                {
                    int len = Database.rawBinaryFactory.pack(field.GetValue(obj), data, offs + 4);
                    Protocol.pack4(data, offs, len);
                    return offs + len + 4;
                }
				
                case t_date: 
                    Protocol.packDate(data, offs, (DateTime)field.GetValue(obj));
                    break;
				
            }
            return offs + size;
        }
		
        internal int packRef(Persistent obj, byte[] data, int offs)
        {
            ObjectHeader hdr = obj.hdr;
            Persistent po = (Persistent) field.GetValue(obj);
            if (po != null)
            {
                ObjectHeader p = po._hdr();
                if (p.opid == 0)
                {
                    p.metaobject.makePersistent(p, hdr);
                }
                else
                {
                    System.Diagnostics.Debug.Assert(hdr.storage.database == p.storage.database);
                }
                Protocol.pack2(data, offs + Protocol.ref_sid_offs, p.storage.id);
                Protocol.pack4(data, offs + Protocol.ref_opid_offs, p.opid);
            }
            else
            {
                for (int i = 0; i < Protocol.ref_size; i++)
                    data[offs + i] = 0;
            }
            return offs + Protocol.ref_size;
        }
		
        internal int packVarying(Persistent obj, byte[] data, int offs)
        {
			
            int i;
            switch (appType)
            {
                case t_sbyte: 
                {
                    sbyte[] arr = (sbyte[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++) 
                    {
                        data[offs++] = (byte)arr[i];
                    }
                    break;
                }
				
                case t_ubyte: 
                {
                    byte[] arr = (byte[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++) 
                    {
                        data[offs++] = arr[i];
                    }
                    break;
                }
				
                case t_short: 
                {
                    short[] arr = (short[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack2(data, offs, arr[i]);
                    }
                    break;
                }
				
                case t_ushort: 
                {
                    ushort[] arr = (ushort[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack2(data, offs, arr[i]);
                    }
                    break;
                }
				
                case t_integer: 
                {
                    int[] arr = (int[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack4(data, offs, arr[i]);
                    }
                    break;
                }
				
                case t_uinteger: 
                {
                    uint[] arr = (uint[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack4(data, offs, (int)arr[i]);
                    }
                    break;
                }
				
                case t_long: 
                {
                    long[] arr = (long[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack8(data, offs, arr[i]);
                    }
                    break;
                }
				
                case t_ulong: 
                {
                    ulong[] arr = (ulong[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack8(data, offs, (long)arr[i]);
                    }
                    break;
                }
				
                case t_float: 
                {
                    float[] arr = (float[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack4(data, offs, 
                            BitConverter.ToInt32(BitConverter.GetBytes(arr[i]), 0));    
                    }
                    break;
                }
				
                case t_double: 
                {
                    double[] arr = (double[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack8(data, offs, BitConverter.DoubleToInt64Bits(arr[i]));
                    }
                    break;
                }
				
                case t_char: 
                {
                    char[] arr = (char[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.pack2(data, offs, arr[i]);
                    }
                    break;
                }
				
                case t_boolean: 
                {
                    bool[] arr = (bool[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        data[offs++] = (byte) (arr[i]?1:0);
                    }
                    break;
                }
				
                case t_date: 
                {
                    DateTime[] arr = (DateTime[]) field.GetValue(obj);
                    for (i = 0; i < arr.Length; i++)
                    {
                        offs = Protocol.packDate(data, offs, arr[i]);
                    }
                    break;
                }
				
            }
            return offs;
        }
		
        internal int packObjectArray(Persistent obj, byte[] data, int offs)
        {
            Persistent[] arr = null;
            ObjectHeader hdr = obj.hdr;
            try
            {
                arr = (Persistent[]) field.GetValue(obj);
            }
            catch (Exception x)
            {
                throw new ApplicationException(x.Message);
            }
            for (int i = 0; i < arr.Length; i++)
            {
                if (arr[i] != null)
                {
                    ObjectHeader p = arr[i]._hdr();
                    if (p.opid == 0)
                    {
                        p.metaobject.makePersistent(p, hdr);
                    }
                    else
                    {
                        System.Diagnostics.Debug.Assert(hdr.storage.database == p.storage.database);
                    }
                    Protocol.pack2(data, offs + Protocol.ref_sid_offs, p.storage.id);
                    Protocol.pack4(data, offs + Protocol.ref_opid_offs, p.opid);
                    offs += Protocol.ref_size;
                }
                else
                {
                    for (int j = Protocol.ref_size; --j >= 0; )
                    {
                        data[offs++] = (byte) 0;
                    }
                }
            }
            return offs;
        }
		
        internal long unpackSignedInteger(byte[] data, int offs)
        {
            switch (size)
            {
                case 1: 
                    return data[offs];
				
                case 2: 
                    return Protocol.unpack2(data, offs);
				
                case 4: 
                    return Protocol.unpack4(data, offs);
				
                default: 
                    return Protocol.unpack8(data, offs);
				
            }
        }
		
        internal long unpackUnsignedInteger(byte[] data, int offs)
        {
            switch (size)
            {
                case 1: 
                    return data[offs] & 0xFF;
				
                case 2: 
                    return Protocol.unpack2(data, offs) & 0xFFFF;
				
                case 4: 
                    return (long) Protocol.unpack4(data, offs) & (int) (- (0x100000000 - 0xFFFFFFFFL));
				
                default: 
                    return Protocol.unpack8(data, offs);
				
            }
        }
		
        internal double unpackReal(byte[] data, int offs)
        {
            if (size == 4)
            {
                return BitConverter.ToSingle(BitConverter.GetBytes(Protocol.unpack4(data, offs)), 0);
            } 
            else 
            { 
                return BitConverter.Int64BitsToDouble(Protocol.unpack8(data, offs));
            }
        }
		
        internal void  storeInteger(Persistent obj, long val)
        {
            switch (appType)
            {
                case t_sbyte: 
                    field.SetValue(obj, (sbyte) val);
                    break;
				
                case t_ubyte: 
                    field.SetValue(obj, (byte) val);
                    break;
				
                case t_short: 
                    field.SetValue(obj, (short) val);
                    break;
				
                case t_ushort: 
                    field.SetValue(obj, (ushort) val);
                    break;
				
                case t_integer: 
                    field.SetValue(obj, (int) val);
                    break;
				
                case t_uinteger: 
                    field.SetValue(obj, (uint) val);
                    break;
				
                case t_long: 
                    field.SetValue(obj, (long)val);
                    break;
				
                case t_ulong: 
                    field.SetValue(obj, (ulong)val);
                    break;
				
                case t_float: 
                    field.SetValue(obj, (float) val);
                    break;
				
                case t_double: 
                    field.SetValue(obj, (double) val);
                    break;
				
                case t_char: 
                    field.SetValue(obj, (char) val);
                    break;
				
                case t_boolean: 
                    field.SetValue(obj, val != 0);
                    break;
				
                case t_date: 
                    field.SetValue(obj, Protocol.NewDateTime(val));
                    break;
				
            }
        }
		
        internal void  storeReal(Persistent obj, double val)
        {
            switch (appType)
            {
                case t_sbyte: 
                    field.SetValue(obj, (sbyte) val);
                    break;
				
                case t_ubyte: 
                    field.SetValue(obj, (byte) val);
                    break;
				
                case t_short: 
                    field.SetValue(obj, (short) val);
                    break;
				
                case t_ushort: 
                    field.SetValue(obj, (ushort) val);
                    break;
				
                case t_integer: 
                    field.SetValue(obj, (int) val);
                    break;
				
                case t_uinteger: 
                    field.SetValue(obj, (uint) val);
                    break;
				
                case t_long: 
                    field.SetValue(obj, (long) val);
                    break;
				
                case t_ulong: 
                    field.SetValue(obj, (ulong) val);
                    break;
				
                case t_float: 
                    field.SetValue(obj, (float) val);
                    break;
				
                case t_double: 
                    field.SetValue(obj, val);
                    break;
				
                case t_char: 
                    field.SetValue(obj, (char) val);
                    break;
				
                case t_boolean: 
                    field.SetValue(obj, val != 0.0);
                    break;
				
            }
        }
		
        internal int storeString(Persistent obj, byte[] data, int offs)
        {
            int len = Protocol.unpack2(data, offs) & 0xFFFF;
            offs += 2;
            if (len != nullString)
            {
                char[] arr = new char[len];
                for (int i = 0; i < len; i++)
                {
                    arr[i] = (char) Protocol.unpack2(data, offs);
                    offs += 2;
                }
                if (field != null)
                {
                    field.SetValue(obj, new String(arr));
                }
            }
            else if (field != null)
            {
                field.SetValue(obj, null);
            }
            return offs;
        }
		
        internal int storeRawBinary(Persistent obj, byte[] data, int offs)
        {
            int size = Protocol.unpack4(data, offs);
            offs += 4;
            if (field != null)
            {
                field.SetValue(obj, Database.rawBinaryFactory.unpack(field.FieldType, data, offs, size));
            }
            return offs + size;
        }
		
		
        internal int unpack(byte[] data, int offs, Persistent obj)
        {
            switch (dbsType)
            {
                case Protocol.fld_signed_integer: 
                    if (field != null)
                    {
                        storeInteger(obj, unpackSignedInteger(data, offs));
                    }
                    break;
				
                case Protocol.fld_unsigned_integer: 
                    if (field != null)
                    {
                        storeInteger(obj, unpackUnsignedInteger(data, offs));
                    }
                    break;
				
                case Protocol.fld_real: 
                    if (field != null)
                    {
                        storeReal(obj, unpackReal(data, offs));
                    }
                    break;
				
                case Protocol.fld_string: 
                    return storeString(obj, data, offs);
				
                case Protocol.fld_raw_binary: 
                    return storeRawBinary(obj, data, offs);
				
            }
            return offs + size;
        }
		
        internal int unpackVarying(byte[] data, int offs, Persistent obj, int varyingLength)
        {
            int i;
			
            if (field == null)
            {
                return offs + size * varyingLength;
            }
            switch (appType)
            {
                case t_sbyte: 
                {
                    sbyte[] arr = new sbyte[varyingLength];
                    if (size == 1)
                    {
                        for (i = 0; i < varyingLength; i++) 
                        { 
                            data[offs++] = (byte)arr[i];
                        }
                    }
                    else
                    {
                        switch (dbsType)
                        {
                            case Protocol.fld_signed_integer: 
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (sbyte) unpackSignedInteger(data, offs);
                                    offs += size;
                                }
                                break;
								
                            case Protocol.fld_unsigned_integer: 
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (sbyte) unpackUnsignedInteger(data, offs);
                                    offs += size;
                                }
                                break;
								
                            case Protocol.fld_real: 
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (sbyte) unpackReal(data, offs);
                                    offs += size;
                                }
                                break;
								
                        }
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_ubyte: 
                {
                    byte[] arr = new byte[varyingLength];
                    if (size == 1)
                    {
                        for (i = 0; i < varyingLength; i++) 
                        { 
                            data[offs++] = arr[i];
                        }
                    }
                    else
                    {
                        switch (dbsType)
                        {
                            case Protocol.fld_signed_integer: 
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (byte) unpackSignedInteger(data, offs);
                                    offs += size;
                                }
                                break;
								
                            case Protocol.fld_unsigned_integer: 
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (byte) unpackUnsignedInteger(data, offs);
                                    offs += size;
                                }
                                break;
								
                            case Protocol.fld_real: 
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (byte) unpackReal(data, offs);
                                    offs += size;
                                }
                                break;
								
                        }
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_short: 
                {
                    short[] arr = new short[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (short) unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (short) unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (short) unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_ushort: 
                {
                    ushort[] arr = new ushort[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (ushort) unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (ushort) unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (ushort) unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_integer: 
                {
                    int[] arr = new int[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (int) unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (int) unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (int) unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }

                case t_uinteger: 
                {
                    uint[] arr = new uint[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (uint) unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (uint) unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (uint) unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_long: 
                {
                    long[] arr = new long[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (long) unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }

                case t_ulong: 
                {
                    ulong[] arr = new ulong[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (ulong)unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (ulong)unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (ulong) unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
				
                case t_float: 
                {
                    float[] arr = new float[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (float) unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (float) unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (float) unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_double: 
                {
                    double[] arr = new double[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (double) unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (double) unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_char: 
                {
                    char[] arr = new char[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (char) unpackSignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (char) unpackUnsignedInteger(data, offs);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = (char) unpackReal(data, offs);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_boolean: 
                {
                    bool[] arr = new bool[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = unpackSignedInteger(data, offs) != 0;
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = unpackUnsignedInteger(data, offs) != 0;
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_real: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                arr[i] = unpackReal(data, offs) != 0.0;
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }
				
                case t_date: 
                {
                    DateTime[] arr = new DateTime[varyingLength];
                    switch (dbsType)
                    {
                        case Protocol.fld_signed_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                long val = unpackSignedInteger(data, offs);
                                arr[i] = Protocol.NewDateTime(val);
                                offs += size;
                            }
                            break;
							
                        case Protocol.fld_unsigned_integer: 
                            for (i = 0; i < varyingLength; i++)
                            {
                                long val = unpackUnsignedInteger(data, offs);
                                arr[i] = Protocol.NewDateTime(val);
                                offs += size;
                            }
                            break;
							
                    }
                    field.SetValue(obj, arr);
                    break;
                }				
            }
            return offs;
        }
    }
}