//-< OptimisticMetaobject.java >-------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update:  9-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Metaobject for optimistic scheme of synchronization
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    /// <summary> Metaobject for optimistic scheme of synchronization. Optimistic means we don't 
    /// lock. Multiple Readers, multiple Writers would be another way of putting it. 
    /// 
    /// Another way to say this is of course, "You'd better know what you're doing", and 
    /// then later, "You've been warned". 
    /// On the other hand, if you do know that this is no problem, the benefit is that 
    /// you save the time for the locking (which is a network request in the Pessimistic
    /// case)
    ///
    /// </summary>
    /// <author>  K.A. Knizhnik 
    /// </author>
    /// <version> 1.0
    /// 
    /// </version>
    public class OptimisticMetaobject:BasicMetaobject
    {
		
        /// <summary>  <code>beginWriteAccess</code> is a no-op. That's why it's optimistic.
        ///
        /// </summary>
        /// <param name="obj">a <code>Persistent</code> object to be modified.
        /// 
        /// </param>
        protected internal override void  beginWriteAccess(ObjectHeader obj)
        {
        }
		
		
        /// <summary>  <code>endAccess</code> adds the object to the cache managers transaction
        /// list if the object has been modified.
        ///
        /// </summary>
        /// <param name="obj">a <code>Persistent</code> object that was (or was not) modified.
        /// 
        /// </param>
        protected internal override void  endAccess(ObjectHeader obj)
        {
            if ((obj.state & (ObjectState.TRANSWRITE | ObjectState.DIRTY)) == ObjectState.DIRTY)
            {
                CacheManager mng = CacheManager.GetCacheManager();
                mng.addToTransaction(obj);
                obj.state |= ObjectState.OPTIMISTIC;
            }
        }
		
        /// <summary>  <code>beginReadAccess</code> is a no-op. We're being optimistic, remember.
        ///
        /// </summary>
        /// <param name="obj">a <code>Persistent</code> object to be modified.
        /// 
        /// </param>
        protected internal override void  beginReadAccess(ObjectHeader obj)
        {
        }
    }
}