//-< TESTGC.CXX >----------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     21-Sep-2001  K.A. Knizhnik  * / [] \ *
//                          Last update: 21-Sep-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Test program for GOODS garbage collector
//-------------------------------------------------------------------*--------*

#include "goods.h"
#include "dbscls.h"

USE_GOODS_NAMESPACE

#define N_ITER     10
#define N_TRANS    100
#define N_RECORDS  1000

class record : public object { 
  public: 
    ref<record> next;
    ref<record> prev;
    int         id;

    METACLASS_DECLARATIONS(record, object);

    record(int id) : object(self_class) 
    { 
        this->id = id;
    }

    void insert(record* rec) { 
        rec->next = next;
        rec->prev = NULL;
        if (next.is_nil()) {
            prev = rec;
        } else { 
            modify(next)->prev = rec;
        }
        next = rec;
    }

    boolean is_empty() const { 
        return next.is_nil();
    }

    void reset() { 
        next = NULL;
        prev = NULL;
    }

    int insert(int id, int n) { 
        while (--n >= 0) { 
            insert(new record(id++));
        }
        return id;
    }

    void initialize() const { 
        if (is_abstract_root()) { 
            ref<record> root = this;
            modify(root)->become(new record(0));
        } 
    }

};

field_descriptor& record::describe_components()
{
    return FIELD(id), FIELD(next), FIELD(prev);
}

REGISTER(record, object, pessimistic_repeatable_read_scheme);


int main(int argc, char* argv[]) 
{
    database db;
    int i, j, id;
    task::initialize(task::normal_stack);

    if (db.open("testdb.cfg")) { 
        ref<record> root;
        db.get_root(root);
        root->initialize();

        for (i = 0; i < N_ITER; i++) { 
            printf("Iteration %d\r", i);
            modify(root)->reset();
            for (id = 1, j = 0; j < N_TRANS; j++) {
                id = modify(root)->insert(id, N_RECORDS);
            }
            id = 0;
            for (ref<record> rec = root->prev; !rec.is_nil(); rec = rec->prev) { 
                assert(rec->id == ++id);
            }
        }
        db.close();
        return EXIT_SUCCESS;
    } else { 
        console::output("Failed to open database\n");
        return EXIT_FAILURE;
    }
}




