//-< OptimisticMetaobject.java >-------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update:  9-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Metaobject for optimistic scheme of synchronization
//-------------------------------------------------------------------*--------*

package goodsjpi;

/**
 * Metaobject for optimistic scheme of synchronization. Optimistic means we don't 
 * lock. Multiple Readers, multiple Writers would be another way of putting it. <br>
 * <br>
 * Another way to say this is of course, "You'd better know what you're doing", and 
 * then later, "You've been warned". <br>
 * On the other hand, if you do know that this is no problem, the benefit is that 
 * you save the time for the locking (which is a network request in the Pessimistic
 * case)
 *
 * @author  K.A. Knizhnik 
 * @version 1.0
 */
public class OptimisticMetaobject extends BasicMetaobject { 

  /**
   *  <code>beginWriteAccess</code> is a no-op. That's why it's optimistic.
   *
   * @param obj a <code>Persistent</code> object to be modified.
   */
    public void beginWriteAccess(Persistent obj) {}


  /**
   *  <code>endAccess</code> adds the object to the cache managers transaction
   * list if the object has been modified.
   *
   * @param obj a <code>Persistent</code> object that was (or was not) modified.
   */
    public void endAccess(Persistent obj) {
	if ((obj.state & (Persistent.TRANSWRITE|Persistent.DIRTY)) == 
	    Persistent.DIRTY) 
	{ 
	    CacheManager mng = CacheManager.getCacheManager();
	    mng.addToTransaction(obj);
	    obj.state |= Persistent.OPTIMISTIC;
	}
    }

  /**
   *  <code>beginReadAccess</code> is a no-op. We're being optimistic, remember.
   *
   * @param obj a <code>Persistent</code> object to be modified.
   */
    public void beginReadAccess(Persistent obj) {}
}


