//-< PessimisticRepeatableReadMetaobject.java >----------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update:  9-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Metaobject for pessimistic scheme of synchronization with repeatable read
// isolation level
//-------------------------------------------------------------------*--------*

package goodsjpi;

/**
 *  <code>PessimisticRepeatableReadMetaobject</code>  for pessimistic scheme of
 *  synchronization with repeatable read  isolation level <br>
 * While extending the Multiple Reader, one writer aproach, this strategy obtains
 * shred locks for reading, thus ensuring that no write is going on at the same time.
 * @author  K.A. Knizhnik 
 * @version 1.0
 */
public class PessimisticRepeatableReadMetaobject extends PessimisticMetaobject
{ 
  /**
   * <code>beginReadAccess</code> locks the object in shared mode.
   *
   * @param obj a <code>Persistent</code> object to be read.
   */
    public void beginReadAccess(Persistent obj) {
	if ((obj.state & (Persistent.SLOCKED|Persistent.XLOCKED)) == 0) { 
	    obj.storage.lock(obj, Protocol.lck_shared, Protocol.lckattr_wait);
	    obj.state |= Persistent.SLOCKED;
	} 
    }

  /**
   *  <code>endAccess</code> adds the object to the transaction list, also when the 
   * object was not modified. Thus the (shared) lock is released.
   *
   * @param obj a <code>Persistent</code> object 
   */
    public void endAccess(Persistent obj) { 
	int state = obj.state;
	if ((state & Persistent.TRANSWRITE) == 0 &&
	    (state & (Persistent.DIRTY|Persistent.XLOCKED)) != 0)
	{
	    CacheManager mng = CacheManager.getCacheManager();
	    mng.addToTransaction(obj);
	} 
	else if ((state & (Persistent.TRANSREAD|Persistent.TRANSWRITE|
			   Persistent.SLOCKED)) == Persistent.SLOCKED)
	{
	    Assert.that((state & Persistent.DIRTY) == 0);
	    CacheManager mng = CacheManager.getCacheManager();
	    mng.addToTransaction(obj);
	}	    
    }
}


