package app;
import java.util.HashSet;
import java.util.Set;

import org.dcg.datamgr.DatabaseThreadPoolMgr;
import org.dcg.datamgr.QueryObject;
import org.dcg.datamgr.DataInsertionObject;
import org.dcg.datamgr.databasejobs.DatabaseReadJob;
import org.dcg.datamgr.databasejobs.DatabaseWriteJob;
import org.dcg.datamgr.goodsindex.RootClass;
import org.dcg.gen.ThreadPoolMgrProperties;

import goodsjpi.Database;
import goodslib.PersistentTreeMap;

/**
 * @author dacosta
 */
public class TestDB {
	final boolean BULK = true;
	final boolean NOTBULK = false;
	static DatabaseThreadPoolMgr pool;
	static int nrOfObjects = 0;
	static ThreadPoolMgrProperties props = new ThreadPoolMgrProperties();
    String sDataBase = "./dbsettings/test2.cfg";


	public TestDB(DatabaseThreadPoolMgr pool) {
		System.out.println("Starting the test");

		// Insert a DataInsertionObject into the WriteJob because otherwise it will not fly
		// Just adding the job to the pool will do the trick.
		// The mgr itself will determine where to send the job to (after determining its type)
		pool.addJob(new DatabaseWriteJob(new DataInsertionObject()));

		// Defining query objects
		HashSet qAND = new HashSet();
		HashSet qOR = new HashSet();
		HashSet qNOT = new HashSet();
		QueryObject qStrP = new QueryObject(qAND, qOR, qNOT);

		pool.addJob(new DatabaseReadJob(qStrP));
	}


	public static void setPoolMgrProperties(String[] args) {
		// Set the thread pool properties using the command line args.
		props.setProperty("maxThreads", args[0]);
		props.setProperty("maxWriteThreads", "1"); // !!!
		props.setProperty("minThreads", args[1]);
		props.setProperty("minWriteThreads", "0"); // !!!
		props.setProperty("maxIdleTime", args[2]);
		props.setProperty("databaseConnector", "goods"); // !!!
        String databaseSystem = args[4];
        props.setProperty("service", databaseSystem);
        if (args.length == 6 && args[5].equals("debug")) {
        	props.setProperty("debug", "true");
        } else {
        	props.setProperty("debug", "false");
        }

        nrOfObjects = java.lang.Integer.parseInt(args[3]);
	}

	
	public static void main(String[] args) {
	    String sDataBase = "./dbsettings/test2.cfg";

		if (args.length < 4) {
			System.err.println("Usage: java Test <maxThreads> <minThreads> <maxIdleTime> <nrOfObjects2bCreated> <database> [debug]");
			System.exit(-1);
		}
		
        // Make sure the database is initialized before starting to use it

		// The following two lines are relevant for the use of the ThreadPoolMgr
		setPoolMgrProperties(args);
		pool = new DatabaseThreadPoolMgr(props);

		System.out.println("\nMain() stats request\n********************\n"+ pool.getStats()+ "********************\n\n");
		new TestDB(pool);
	}
	
}
