package org.dcg.datamgr;

import org.dcg.datamgr.goodsindex.RootClass;

import goodsjpi.Database;

/**
 * 
 * @author dacosta
 */
public class DBConnector {
	/**
	 * The name of the service that requires the database
	 */
	private String _serviceName;

	/**
	 * Name of the type/ vendor of the database.
	 * i.e. GOODS, Oracle etc.
	 */
	private String _databaseName;
	
	/**
	 * path to the lacation of the config file of the database
	 */
	private String _databaseLocation;
	
	/**
	 * The config file that controls the database
	 */
	private String _databaseConfigFile;

    /**
     * Indicator to show whether the database for the particular service has been initialized
     */
    private boolean _isDatabaseServiceInitialized = false;
    
	/**
	 * Method DBConnector.
	 * 
	 * @param service Contains the name of the service in full class definition style
	 * @param db Contains the vendorName of the database (GOODS, Oracle, ProgresSQL, etc)
	 * @param dbLocation Location (path) of the network where the config file can be found
	 */
	public DBConnector(String service, String db, String dbLocation) {
		_serviceName = service;
		_databaseLocation = dbLocation;
		_databaseName = db;
		_databaseConfigFile = _databaseLocation + _serviceName +".cfg";
        _isDatabaseServiceInitialized = isDatabaseInitialized();
	}
	
	/**
	 * Returns the _databaseConfigFile.
	 * @return String
	 */
	public String getDatabaseConfigFile() {
		return _databaseConfigFile;
	}

    public boolean isServiceDatabaseInitialized() {
        return _isDatabaseServiceInitialized;
    }

    private boolean isDatabaseInitialized() {
        Database db = new Database();
        db.open(_databaseConfigFile);
        if (db.getRoot() == null) { // The database has NOT been initialized
            db.setRoot(new RootClass());
            RootClass r = (RootClass) db.getRoot();
            db.close();
            r = null;
            db = null;
            return false;
        } else {
            db.close();
            db = null;
            return true;
        }
    }

}
